@@ -29,35 +29,72 @@
  * to add clustering support.
  */
 
+#include "config.h"
+
 #define	MAIN_PROGRAM
 
-#include <cron.h>
+#include <errno.h>
+#include <locale.h>
+#include <pwd.h>
+#include <signal.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <sys/stat.h>
+#include <sys/types.h>
+#include <sys/wait.h>
+#include <unistd.h>
+#include <utime.h>
+
+#ifdef WITH_PAM
+# include <security/pam_appl.h>
+#endif
+
 #ifdef WITH_SELINUX
 # include <selinux/selinux.h>
 # include <selinux/context.h>
-# include <selinux/av_permissions.h>
 #endif
 
+#include "cronie_common.h"
+#include "bitstring.h"
+#include "externs.h"
+#include "funcs.h"
+#include "globals.h"
+#include "macros.h"
+#include "pathnames.h"
+#include "structs.h"
+
 #define NHEADER_LINES 0
 
-enum opt_t {opt_unknown, opt_list, opt_delete, opt_edit, opt_replace, opt_hostset, opt_hostget};
+#define COMMENT_COLOR "\x1B[34;1m"
+#define ERROR_COLOR "\x1B[31;1m"
+#define RESET_COLOR "\x1B[0m"
+
+enum opt_t {
+	opt_unknown, opt_list, opt_delete, opt_edit, opt_replace, opt_hostset,
+	opt_hostget, opt_test
+};
 
 #if DEBUGGING
-static char *Options[] = {"???", "list", "delete", "edit", "replace", "hostset", "hostget"};
+static const char *Options[] = {
+	"???", "list", "delete", "edit", "replace", "hostset", "hostget", "test"
+};
 
 # ifdef WITH_SELINUX
-static char *getoptargs = "u:lerisncx:";
+static const char *getoptargs = "u:lerisncx:VT";
 # else
-static char *getoptargs = "u:lerincx:";
+static const char *getoptargs = "u:lerincx:VT";
 # endif
 #else
 # ifdef WITH_SELINUX
-static char *getoptargs = "u:lerisnc";
+static const char *getoptargs = "u:lerisncVT";
 # else
-static char *getoptargs = "u:lerinc";
+static const char *getoptargs = "u:lerincVT";
 # endif
 #endif
+#ifdef WITH_SELINUX
 static char *selinux_context = 0;
+#endif
 
 static PID_T Pid;
 static char User[MAX_UNAME], RealUser[MAX_UNAME];
@@ -73,35 +110,53 @@ static void list_cmd(void),
 delete_cmd(void),
 edit_cmd(void),
 poke_daemon(void),
-check_error(const char *), parse_args(int c, char *v[]), die(int);
-static int replace_cmd(void), hostset_cmd(void), hostget_cmd(void);
-static char *host_specific_filename(char *filename, int prefix);
-static char *tmp_path(void);
+check_error(const char *), parse_args(int c, char *v[]),
+die(int) ATTRIBUTE_NORETURN;
+static int replace_cmd(void), hostset_cmd(void), hostget_cmd(void),
+test_cmd(void), check_syntax(FILE *);
+static char *host_specific_filename(const char *prefix,
+const char *suffix);
+static const char *tmp_path(void);
 
+static void usage(const char *msg) ATTRIBUTE_NORETURN;
 static void usage(const char *msg) {
 	fprintf(stderr, "%s: usage error: %s\n", ProgramName, msg);
-	fprintf(stderr, "usage:\t%s [-u user] file\n", ProgramName);
-	fprintf(stderr, "\t%s [-u user] [ -e | -l | -r ]\n", ProgramName);
-	fprintf(stderr, "\t%s -n [ hostname ]\n", ProgramName);
-	fprintf(stderr, "\t%s -c\n", ProgramName);
-	fprintf(stderr, "\t\t(default operation is replace, per 1003.2)\n");
-	fprintf(stderr, "\t-e\t(edit user's crontab)\n");
-	fprintf(stderr, "\t-l\t(list user's crontab)\n");
-	fprintf(stderr, "\t-r\t(delete user's crontab)\n");
-	fprintf(stderr, "\t-i\t(prompt before deleting user's crontab)\n");
-	fprintf(stderr, "\t-n\t(set host in cluster to run users' crontabs)\n");
-	fprintf(stderr, "\t-c\t(get host in cluster to run users' crontabs)\n");
+	fprintf(stderr, "Usage:\n");
+	fprintf(stderr, " %s [options] file\n", ProgramName);
+	fprintf(stderr, " %s [options]\n", ProgramName);
+	fprintf(stderr, " %s -n [hostname]\n", ProgramName);
+	fprintf(stderr, "\n");
+	fprintf(stderr, "Options:\n");
+	fprintf(stderr, " -u <user>  define user\n");
+	fprintf(stderr, " -e         edit user's crontab\n");
+	fprintf(stderr, " -l         list user's crontab\n");
+	fprintf(stderr, " -r         delete user's crontab\n");
+	fprintf(stderr, " -i         prompt before deleting\n");
+	fprintf(stderr, " -n <host>  set host in cluster to run users' crontabs\n");
+	fprintf(stderr, " -c         get host in cluster to run users' crontabs\n");
+	fprintf(stderr, " -T <file>  test a crontab file syntax\n");
 #ifdef WITH_SELINUX
-	fprintf(stderr, "\t-s\t(selinux context)\n");
+	fprintf(stderr, " -s         selinux context\n");
 #endif
+	fprintf(stderr, " -V         print version and exit\n");
+#ifdef DEBUGGING
+	fprintf(stderr, " -x <mask>  enable debugging\n");
+#endif
+	fprintf(stderr, "\nDefault operation is replace, per 1003.2\n");
 	exit(ERROR_EXIT);
 }
 
 int main(int argc, char *argv[]) {
 	int exitstatus;
 
+	if ((ProgramName = strrchr(argv[0], '/')) == NULL) {
+		ProgramName = argv[0];
+	}
+	else {
+		++ProgramName;
+	}
+
 	Pid = getpid();
-	ProgramName = argv[0];
 	MailCmd[0] = '\0';
 	cron_default_mail_charset[0] = '\0';
 
@@ -110,11 +165,6 @@ int main(int argc, char *argv[]) {
 #if defined(BSD)
 	setlinebuf(stderr);
 #endif
-	char *n = "-";	/*set the n string to - so we have a valid string to use */
-	/*should we desire to make changes to behavior later. */
-	if (argv[1] == NULL) {	/* change behavior to allow crontab to take stdin with no '-' */
-		argv[1] = n;
-	}
 	parse_args(argc, argv);	/* sets many globals, opens a file */
 	check_spool_dir();
 	if (!allowed(RealUser, CRON_ALLOW, CRON_DENY)) {
@@ -127,7 +177,7 @@ int main(int argc, char *argv[]) {
 	}
 
 #if defined(WITH_PAM)
-	if (cron_start_pam(pw) != PAM_SUCCESS) {
+	if (getuid() != 0 && cron_start_pam(pw) != PAM_SUCCESS) {
 		fprintf(stderr,
 			"You (%s) are not allowed to access to (%s) because of pam configuration.\n",
 			User, ProgramName);
@@ -161,10 +211,16 @@ int main(int argc, char *argv[]) {
 		if (hostget_cmd() < 0)
 			exitstatus = ERROR_EXIT;
 		break;
+	case opt_test:
+		if (test_cmd() < 0)
+			exitstatus = ERROR_EXIT;
+		break;
 	default:
 		abort();
 	}
+#ifdef WITH_PAM
 	cron_close_pam();
+#endif
 	exit(exitstatus);
  /*NOTREACHED*/}
 
@@ -200,16 +256,16 @@ static void parse_args(int argc, char *argv[]) {
 				fprintf(stderr, "must be privileged to use -u\n");
 				exit(ERROR_EXIT);
 			}
-
+#ifdef WITH_SELINUX
 			if (crontab_security_access() != 0) {
 				fprintf(stderr,
 					"Access denied by SELinux, must be privileged to use -u\n");
 				exit(ERROR_EXIT);
 			}
-
-			if (Option == opt_hostset || Option == opt_hostget) {
-				fprintf(stderr,
-					"cannot use -u with -n or -c\n");
+#endif
+			if (Option == opt_hostset || Option == opt_hostget ||
+				Option == opt_test) {
+				fprintf(stderr, "cannot use -u with -n, -c or -T\n");
 				exit(ERROR_EXIT);
 			}
 
@@ -237,6 +293,11 @@ static void parse_args(int argc, char *argv[]) {
 				usage("only one operation permitted");
 			Option = opt_edit;
 			break;
+		case 'T':
+			if (Option != opt_unknown)
+				usage("only one operation permitted");
+			Option = opt_test;
+			break;
 		case 'i':
 			PromptOnDelete = 1;
 			break;
@@ -250,15 +311,13 @@ static void parse_args(int argc, char *argv[]) {
 #endif
 		case 'n':
 			if (MY_UID(pw) != ROOT_UID) {
-				fprintf(stderr,
-					"must be privileged to set host with -n\n");
+				fprintf(stderr, "must be privileged to set host with -n\n");
 				exit(ERROR_EXIT);
 			}
 			if (Option != opt_unknown)
 				usage("only one operation permitted");
 			if (strcmp(User, RealUser) != 0) {
-				fprintf(stderr,
-					"cannot use -u with -n or -c\n");
+				fprintf(stderr, "cannot use -u with -n or -c\n");
 				exit(ERROR_EXIT);
 			}
 			Option = opt_hostset;
@@ -267,12 +326,14 @@ static void parse_args(int argc, char *argv[]) {
 			if (Option != opt_unknown)
 				usage("only one operation permitted");
 			if (strcmp(User, RealUser) != 0) {
-				fprintf(stderr,
-					"cannot use -u with -n or -c\n");
+				fprintf(stderr, "cannot use -u with -n or -c\n");
 				exit(ERROR_EXIT);
 			}
 			Option = opt_hostget;
 			break;
+		case 'V':
+			puts(PACKAGE_STRING);
+			exit(EXIT_SUCCESS);
 		default:
 			usage("unrecognized option");
 		}
@@ -280,7 +341,7 @@ static void parse_args(int argc, char *argv[]) {
 
 	endpwent();
 
-	if (Option == opt_hostset && argv[optind] != NULL) {            
+	if (Option == opt_hostset && argv[optind] != NULL) {
 		HostSpecified = 1;
 		if (strlen(argv[optind]) >= sizeof Host)
 			usage("hostname too long");
@@ -288,22 +349,31 @@ static void parse_args(int argc, char *argv[]) {
 		optind++;
 	}
 
-	if (Option != opt_unknown) {
-		if (argv[optind] != NULL)
-			usage("no arguments permitted after this option");
+	if (Option == opt_unknown) {
+		/* replace is the default option */
+		Option = opt_replace;
 	}
-	else {
+
+	if (Option == opt_replace || Option == opt_test) {
 		if (argv[optind] != NULL) {
-			Option = opt_replace;
 			if (strlen(argv[optind]) >= sizeof Filename)
 				usage("filename too long");
 			(void) strcpy(Filename, argv[optind]);
+			optind++;
+		}
+		else if (isatty(STDIN_FILENO)) {
+			usage("file name or - (for stdin) must be specified");
+		}
+		else {
+			strcpy(Filename, "-");
 		}
-		else
-			usage("file name must be specified for replace");
 	}
 
-	if (Option == opt_replace) {
+	if (Option != opt_unknown && argv[optind] != NULL) {
+		usage("no arguments permitted after this option");
+	}
+
+	if (Filename[0] != '\0') {
 		if (!strcmp(Filename, "-"))
 			NewCrontab = stdin;
 		else {
@@ -314,6 +384,7 @@ static void parse_args(int argc, char *argv[]) {
 			 * Arnt Gulbrandsen <agulbra@pvv.unit.no> for spotting
 			 * the race.
 			 */
+			struct stat sb;
 
 			if (swap_uids() < OK) {
 				perror("swapping uids");
@@ -323,6 +394,16 @@ static void parse_args(int argc, char *argv[]) {
 				perror(Filename);
 				exit(ERROR_EXIT);
 			}
+			if (fstat(fileno(NewCrontab), &sb) < 0) {
+				perror(Filename);
+				exit(ERROR_EXIT);
+			}
+			if ((sb.st_mode & S_IFMT) == S_IFDIR) {
+				fprintf(stderr,
+					"invalid crontab file: '%s' is a directory\n", Filename);
+				fclose(NewCrontab);
+				exit(ERROR_EXIT);
+			}
 			if (swap_uids_back() < OK) {
 				perror("swapping uids back");
 				exit(ERROR_EXIT);
@@ -331,13 +412,16 @@ static void parse_args(int argc, char *argv[]) {
 	}
 
 	Debug(DMISC, ("user=%s, file=%s, option=%s\n",
-			User, Filename, Options[(int) Option]))
+			User, Filename, Options[(int) Option]));
 }
 
 static void list_cmd(void) {
 	char n[MAX_FNAME];
 	FILE *f;
 	int ch;
+	const int colorize = isatty(STDOUT) && getenv("NO_COLOR") == NULL;
+	int new_line = 1;
+	int in_comment = 0;
 
 	log_it(RealUser, Pid, "LIST", User, 0);
 	if (!glue_strings(n, sizeof n, SPOOL_DIR, User, '/')) {
@@ -354,9 +438,28 @@ static void list_cmd(void) {
 
 	/* file is open. copy to stdout, close.
 	 */
-	Set_LineNum(1)
-		while (EOF != (ch = get_char(f)))
+	Set_LineNum(1);
+	while (EOF != (ch = get_char(f))) {
+		if (colorize) {
+			if (!in_comment && new_line && ch == '#') {
+				in_comment = 1;
+				fputs(COMMENT_COLOR, stdout);
+			}
+			if (in_comment && ch == '\n') {
+				in_comment = 0;
+				fputs(RESET_COLOR, stdout);
+			}
+		}
 		putchar(ch);
+		new_line = ch == '\n';
+	}
+	/* no new line at EOF */
+	if (colorize && !new_line) {
+		putchar('\n');
+		fputs(ERROR_COLOR "No end-of-line character at the end of file"
+			RESET_COLOR, stdout);
+		putchar('\n');
+	}
 	fclose(f);
 }
 
@@ -365,7 +468,7 @@ static void delete_cmd(void) {
 	if (PromptOnDelete == 1) {
 		printf("crontab: really delete %s's crontab? ", User);
 		fflush(stdout);
-		if ((fgets(n, MAX_FNAME - 1, stdin) == 0L)
+		if ((fgets(n, MAX_FNAME - 1, stdin) == NULL)
 			|| ((n[0] != 'Y') && (n[0] != 'y'))
 			)
 			exit(0);
@@ -391,8 +494,8 @@ static void check_error(const char *msg) {
 	fprintf(stderr, "\"%s\":%d: %s\n", Filename, LineNumber - 1, msg);
 }
 
-static char *tmp_path() {
-	char *tmpdir = NULL;
+static const char *tmp_path(void) {
+	const char *tmpdir = NULL;
 
 	if ((getuid() == geteuid()) && (getgid() == getegid())) {
 		tmpdir = getenv("TMPDIR");
@@ -400,26 +503,26 @@ static char *tmp_path() {
 	return tmpdir ? tmpdir : "/tmp";
 }
 
-static char *host_specific_filename(char *filename, int prefix)
-{
+static char *host_specific_filename(const char *prefix, const char *suffix) {
 	/*
 	 * For cluster-wide use, where there is otherwise risk of the same
-	 * name being generated on more than one host at once, prefix with
-	 * "hostname." or suffix with ".hostname" as requested, and return
-	 * static buffer or NULL on failure.
+	 * name being generated on more than one host at once, insert hostname
+	 * separated with dots, and return static buffer or NULL on failure.
 	 */
 
 	static char safename[MAX_FNAME];
-	char hostname[MAXHOSTNAMELEN];
+	char hostname[MAX_FNAME];
 
 	if (gethostname(hostname, sizeof hostname) != 0)
 		return NULL;
 
 	if (prefix) {
-		if (!glue_strings(safename, sizeof safename, hostname, filename, '.'))
+		if (!glue_strings(safename, sizeof safename, prefix, hostname, '.'))
 			return NULL;
-	} else {
-		if (!glue_strings(safename, sizeof safename, filename, hostname, '.'))
+		strcpy(hostname, safename);
+	}
+	if (suffix) {
+		if (!glue_strings(safename, sizeof safename, hostname, suffix, '.'))
 			return NULL;
 	}
 
@@ -427,7 +530,8 @@ static char *host_specific_filename(char *filename, int prefix)
 }
 
 static void edit_cmd(void) {
-	char n[MAX_FNAME], q[MAX_TEMPSTR], *editor;
+	char n[MAX_FNAME], q[MAX_TEMPSTR];
+	const char *editor;
 	FILE *f;
 	int ch = '\0', t;
 	struct stat statbuf;
@@ -480,14 +584,14 @@ static void edit_cmd(void) {
 		goto fatal;
 	}
 
-	Set_LineNum(1)
-		/* 
-		 * NHEADER_LINES processing removed for clarity
-		 * (NHEADER_LINES == 0 in all Red Hat crontabs)
-		 */
-		/* copy the rest of the crontab (if any) to the temp file.
-		 */
-		if (EOF != ch)
+	Set_LineNum(1);
+	/* 
+	 * NHEADER_LINES processing removed for clarity
+	 * (NHEADER_LINES == 0 in all Red Hat crontabs)
+	 */
+	/* copy the rest of the crontab (if any) to the temp file.
+	 */
+	if (EOF != ch)
 		while (EOF != (ch = get_char(f)))
 			putc(ch, NewCrontab);
 
@@ -518,10 +622,10 @@ static void edit_cmd(void) {
 		perror(Filename);
 		exit(ERROR_EXIT);
 	}
-        if (swap_uids() == -1) {
-                perror("swapping uids");
-                exit(ERROR_EXIT);
-        }
+	if (swap_uids() == -1) {
+		perror("swapping uids");
+		exit(ERROR_EXIT);
+	}
 	/* Set it to 1970 */
 	utimebuf.actime = 0;
 	utimebuf.modtime = 0;
@@ -644,7 +748,7 @@ static void edit_cmd(void) {
 		perror("swapping uids back");
 		exit(ERROR_EXIT);
 	}
-	if (NewCrontab == 0L) {
+	if (NewCrontab == NULL) {
 		perror("fopen");
 		goto fatal;
 	}
@@ -653,10 +757,10 @@ static void edit_cmd(void) {
 		break;
 	case -1:
 		for (;;) {
-			printf("Do you want to retry the same edit? ");
+			printf("Do you want to retry the same edit? (Y/N) ");
 			fflush(stdout);
 			q[0] = '\0';
-			if (fgets(q, sizeof q, stdin) == 0L)
+			if (fgets(q, sizeof q, stdin) == NULL)
 				continue;
 			switch (q[0]) {
 			case 'y':
@@ -684,22 +788,33 @@ static void edit_cmd(void) {
 	log_it(RealUser, Pid, "END EDIT", User, 0);
 }
 
+/*
+* Check if crontab file can be installed or not 
+*/
+static int test_cmd(void) {
+	if (check_syntax(NewCrontab) < 0) {
+		fprintf(stderr, "Invalid crontab file. Syntax issues were found.\n");
+		return (-2);
+	}
+	else {
+		fprintf(stderr, "No syntax issues were found in the crontab file.\n");
+	}
+	return (0);
+}
+
 /* returns	0	on success
  *		-1	on syntax error
  *		-2	on install error
  */
 static int replace_cmd(void) {
-	char n[MAX_FNAME], envstr[MAX_ENVSTR];
+	char n[MAX_FNAME];
 	FILE *tmp;
-	int ch, eof, fd;
+	int ch, fd;
 	int error = 0;
-	entry *e;
 	uid_t file_owner;
-	char **envp;
 	char *safename;
 
-
-	safename = host_specific_filename("tmp.XXXXXXXXXX", 1);
+	safename = host_specific_filename("#tmp", "XXXXXXXXXX");
 	if (!safename || !glue_strings(TempFilename, sizeof TempFilename, SPOOL_DIR,
 			safename, '/')) {
 		TempFilename[0] = '\0';
@@ -736,8 +851,8 @@ static int replace_cmd(void) {
 	/* copy the crontab to the tmp
 	 */
 	rewind(NewCrontab);
-	Set_LineNum(1)
-		while (EOF != (ch = get_char(NewCrontab)))
+	Set_LineNum(1);
+	while (EOF != (ch = get_char(NewCrontab)))
 		putc(ch, tmp);
 	if (ftruncate(fileno(tmp), ftell(tmp)) == -1) {
 		fprintf(stderr, "%s: error while writing new crontab to %s\n",
@@ -746,74 +861,33 @@ static int replace_cmd(void) {
 		error = -2;
 		goto done;
 	}
-	fflush(tmp);
-	rewind(tmp);
-	if (ferror(tmp)) {
+	if (ferror(tmp) || fflush(tmp) || fsync(fileno(tmp))) {
 		fprintf(stderr, "%s: error while writing new crontab to %s\n",
 			ProgramName, TempFilename);
 		fclose(tmp);
 		error = -2;
 		goto done;
 	}
+	rewind(tmp);
 
-	/* check the syntax of the file being installed.
-	 */
-
-	/* BUG: was reporting errors after the EOF if there were any errors
-	 * in the file proper -- kludged it by stopping after first error.
-	 *      vix 31mar87
-	 */
-	Set_LineNum(1 - NHEADER_LINES)
-		CheckErrorCount = 0;
-	eof = FALSE;
-
-	envp = env_init();
-	if (envp == NULL) {
-		fprintf(stderr, "%s: Cannot allocate memory.\n", ProgramName);
+	if ((error = check_syntax(tmp)) < 0) {
+		fprintf(stderr, "Invalid crontab file, can't install.\n");
 		fclose(tmp);
-		error = -2;
 		goto done;
 	}
 
-	while (!CheckErrorCount && !eof) {
-		switch (load_env(envstr, tmp)) {
-		case ERR:
-			/* check for data before the EOF */
-			if (envstr[0] != '\0') {
-				Set_LineNum(LineNumber + 1);
-				check_error("premature EOF");
-			}
-			eof = TRUE;
-			break;
-		case FALSE:
-			e = load_entry(tmp, check_error, pw, envp);
-			if (e)
-				free_entry(e);
-			break;
-		case TRUE:
-			break;
-		}
-	}
-	env_free(envp);
-
-	if (CheckErrorCount != 0) {
-		fprintf(stderr, "errors in crontab file, can't install.\n");
-		fclose(tmp);
-		error = -1;
-		goto done;
-	}
-
-	file_owner = (getgid() == getegid())? ROOT_UID : pw->pw_uid;
+	file_owner = (getgid() == geteuid() &&
+		getgid() == getegid()) ? ROOT_UID : pw->pw_uid;
 
 #ifdef HAVE_FCHOWN
-	if (fchown(fileno(tmp), file_owner, -1) < OK) {
+	if (fchown(fileno(tmp), file_owner, (gid_t)-1) < OK) {
 		perror("fchown");
 		fclose(tmp);
 		error = -2;
 		goto done;
 	}
 #else
-	if (chown(TempFilename, file_owner, -1) < OK) {
+	if (chown(TempFilename, file_owner, (gid_t)-1) < OK) {
 		perror("chown");
 		fclose(tmp);
 		error = -2;
@@ -855,19 +929,93 @@ static int replace_cmd(void) {
 	return (error);
 }
 
+/*
+ * Check the syntax of a crontab file
+ * Returns:
+ *   0 no syntax issues
+ *  -1 syntax issue (can be fixed by user)
+ *  -2 any other error, which can not be fixed by user
+ */
+static int check_syntax(FILE * crontab_file) {
+	char **envp = env_init();
+	int eof = FALSE;
+	int envs = 0, entries = 0;
+	
+	CheckErrorCount = 0;
+	Set_LineNum(1 - NHEADER_LINES);
+
+	if (envp == NULL) {
+		fprintf(stderr, "%s: Cannot allocate memory.\n", ProgramName);
+		return (-2);
+	}
+
+	while (!CheckErrorCount && !eof) {
+		char envstr[MAX_ENVSTR];
+		entry *e;
+
+		if (!skip_comments(crontab_file)) {
+			check_error
+				("too much non-parseable content (comments, empty lines, spaces)");
+			break;
+		}
+
+		switch (load_env(envstr, crontab_file)) {
+		case ERR:
+			/* check for data before the EOF */
+			if (envstr[0] != '\0') {
+				Set_LineNum(LineNumber + 1);
+				check_error("premature EOF");
+			}
+			eof = TRUE;
+			break;
+		case FALSE:
+			e = load_entry(crontab_file, check_error, pw, envp);
+			if (e) {
+				++entries;
+				free_entry(e);
+			}
+			break;
+		case TRUE:
+			++envs;
+			break;
+		}
+	}
+	env_free(envp);
+
+	if (envs > MAX_USER_ENVS) {
+		fprintf(stderr,
+			"There are too many environment variables in the crontab file. Limit: %d\n",
+			MAX_USER_ENVS);
+		return (-1);
+	}
+
+	if (entries > MAX_USER_ENTRIES) {
+		fprintf(stderr,
+			"There are too many entries in the crontab file. Limit: %d\n",
+			MAX_USER_ENTRIES);
+		return (-1);
+	}
+
+	if (CheckErrorCount != 0) {
+		return (-1);
+	}
+
+	return 0;
+}
+
 static int hostset_cmd(void) {
 	char n[MAX_FNAME];
 	FILE *tmp;
 	int fd;
 	int error = 0;
 	char *safename;
-	
+
 	if (!HostSpecified)
 		gethostname(Host, sizeof Host);
-	
-	safename = host_specific_filename("tmp.XXXXXXXXXX", 1);
+
+	safename = host_specific_filename("#tmp", "XXXXXXXXXX");
 	if (!safename || !glue_strings(TempFilename, sizeof TempFilename, SPOOL_DIR,
-		safename, '/')) {
+			safename, '/')) {
 		TempFilename[0] = '\0';
 		fprintf(stderr, "path too long\n");
 		return (-2);
@@ -886,7 +1034,7 @@ static int hostset_cmd(void) {
 	(void) signal(SIGINT, die);
 	(void) signal(SIGQUIT, die);
 
-	(void) fchmod(fd, 0600); /* not all mkstemp() implementations do this */
+	(void) fchmod(fd, 0600);	/* not all mkstemp() implementations do this */
 
 	if (fprintf(tmp, "%s\n", Host) < 0 || fclose(tmp) == EOF) {
 		fprintf(stderr, "%s: error while writing to %s\n",
@@ -938,7 +1086,7 @@ static int hostget_cmd(void) {
 			fprintf(stderr, "File %s not found\n", n);
 		else
 			perror(n);
-			return (-2);
+		return (-2);
 	}
 
 	if (get_string(Host, sizeof Host, f, "\n") == EOF) {
@@ -964,7 +1112,7 @@ static void poke_daemon(void) {
 	}
 }
 
-static void die(int x) {
+static void die(int x ATTRIBUTE_UNUSED) {
 	if (TempFilename[0])
 		(void) unlink(TempFilename);
 	_exit(ERROR_EXIT);
